"""
Tools for reliably and robustly extracting versioning info from python objects.
"""

from __future__ import annotations

import dataclasses
import importlib
import sys
from collections.abc import Callable
from types import BuiltinMethodType, ModuleType
from typing import Any, Self, TypeAlias

VersionScraperType: TypeAlias = Callable[[str], str | None]
VersionScrapingMap: TypeAlias = dict[str, VersionScraperType]


@dataclasses.dataclass(frozen=True)
class VersionInfo:
    """
    Immutable record of an object's module, qualified name, and module version.

    This is useful for capturing provenance metadata about classes and instances,
    e.g. for serialization or reproducibility tracking.

    Attributes:
        module: The dotted module path where the object's type is defined.
        qualname: The qualified name of the object's type within its module. A None
            qualname indicates that the object is itself a module.
        version: The version string of the top-level package, or ``None`` if no version
            could be determined.

    Example:
        >>> from pyiron_snippets import versions
        >>>
        >>> versions.VersionInfo.of(42)  # doctest: +ELLIPSIS
        VersionInfo(module='builtins', qualname='int', version=...)

    Note:
        For object instances, this is version info about that object's _type_, not
        information about where the instance itself is living.
    """

    module: str
    qualname: str | None
    version: str | None

    @property
    def fully_qualified_name(self) -> str:
        if self.qualname is None:
            return self.module
        return f"{self.module}.{self.qualname}"

    @classmethod
    def of(
        cls,
        obj: object,
        version_scraping: VersionScrapingMap | None = None,
        forbid_main: bool = False,
        forbid_locals: bool = False,
        require_version: bool = False,
    ) -> VersionInfo:
        """
        Construct a :class:`VersionInfo` by introspecting *obj*.

        If a `__module__` or `__qualname__` is immediately available, they are used,
        otherwise the same fields are sought on the object's type.

        Args:
            obj: The object to introspect.
            version_scraping: Optional mapping from top-level package names to
                callables that return a version string (or ``None``). Used to
                handle packages that don't expose ``__version__``.
            forbid_main: If ``True``, raise :exc:`ValueError` when the module
                is ``__main__``.
            forbid_locals: If ``True``, raise :exc:`ValueError` when the
                qualname contains ``<locals>`` (i.e. the type was defined
                inside a function).
            require_version: If ``True``, raise :exc:`ValueError` when no
                version can be determined for the module.

        Returns:
            A new :class:`VersionInfo` instance.

        Raises:
            ValueError: If any of the ``forbid_*`` / ``require_*`` constraints
                are violated.
        """
        module = get_module(obj)
        qualname = get_qualname(obj)
        version = get_version(module, version_scraping=version_scraping)
        info = cls(module=module, qualname=qualname, version=version)
        info.validate_constraints(
            forbid_main=forbid_main,
            forbid_locals=forbid_locals,
            require_version=require_version,
        )
        return info

    def validate_constraints(
        self,
        forbid_main: bool = False,
        forbid_locals: bool = False,
        require_version: bool = False,
    ) -> Self:
        if forbid_main and "__main__" in self.module:
            raise ValueError(f"Found forbidden module '__main__' in module for {self}")

        if forbid_locals and self.qualname is not None and "<locals>" in self.qualname:
            raise ValueError(f"Found forbidden <locals> in qualname for {self}")

        if require_version and self.version is None:
            raise ValueError(f"Could not find a version for {self}")

        return self


@dataclasses.dataclass(frozen=True)
class VersionInfoFactory:
    """
    A simple stateful wrapper for :class:`VersionInfo` that is useful when getting
    info from multiple objects with the same settings.
    """

    version_scraping: VersionScrapingMap | None = None
    forbid_main: bool = False
    forbid_locals: bool = False
    require_version: bool = False

    def of(self, obj: object) -> VersionInfo:
        return VersionInfo.of(
            obj,
            version_scraping=self.version_scraping,
            forbid_main=self.forbid_main,
            forbid_locals=self.forbid_locals,
            require_version=self.require_version,
        )

    def validate_constraints(self, info: VersionInfo) -> VersionInfo:
        return info.validate_constraints(
            forbid_main=self.forbid_main,
            forbid_locals=self.forbid_locals,
            require_version=self.require_version,
        )


def get_module(obj: Any) -> str:
    """
    Get module information for an arbitrary object.

    Note:
        For object instances, this is version info about that object's _type_, not
        information about where the instance itself is living.
    """
    if isinstance(obj, ModuleType):
        return obj.__name__

    # Try the obvious path first
    module = getattr(obj, "__module__", None)
    if module is not None:
        return module

    # For bound builtin methods, look up the defining class
    if isinstance(obj, BuiltinMethodType):
        # obj.__self__ is the instance, obj.__name__ is the method name
        self_obj = getattr(obj, "__self__", None)
        if self_obj is not None:
            for cls in type(self_obj).__mro__:
                if obj.__name__ in cls.__dict__:
                    module = getattr(cls.__dict__[obj.__name__], "__module__", None)
                    if module is not None:
                        return module
            # Fall back to the type's module
            module = getattr(type(self_obj), "__module__", None)
            if module is not None:
                return module

    # Last resort
    module = getattr(type(obj), "__module__", None)
    if module is not None:
        return module

    raise AttributeError(
        f"Could not find a module on obj {obj} or type(obj) {type(obj)}."
    )


def get_qualname(obj: Any) -> str | None:
    """
    Get module information for an arbitrary object.

    Note:
        For object instances, this is version info about that object's _type_, not
        information about where the instance itself is living.
    """
    if isinstance(obj, ModuleType):
        return None
    qualname = getattr(obj, "__qualname__", getattr(type(obj), "__qualname__", None))
    if not isinstance(qualname, str):
        raise TypeError(f"Expected a string __qualname__, but {obj} had {qualname}.")
    if len(qualname) == 0:
        raise ValueError(f"Expected a non-empty qualname string for {obj}.")
    return qualname


def get_version(
    module_name: str,
    version_scraping: VersionScrapingMap | None = None,
) -> str | None:
    """
    Given a module name, get its associated version (if any) by iteratively checking
    each module level for an available version. By default, this simply looks for the
    :attr:`__version__` attribute on the imported module, but searching behaviour can
    be customized with the :arg:`version_scraping` argument.

    The first found version walking up the module path takes precedence over higher
    versions, and the version scraping map entries take precedence over the default
    :attr:`__version__` check at each step while walking.

    For :mod:`builtins`, the python interpreter version is given.

    Args:
        module_name (str): The module to recursively examine.
        version_scraping (VersionScrapingMap | None): Since some modules may store
            their version in other ways, this provides an optional map between module
            names and callables to leverage for extracting that module's version.

    Returns:
        (str | None): The module's version as a string, if any can be found.

    Warnings:
        This imports the module in the process, so it is not "safe".
    """
    if module_name == "builtins":
        return _python_version()

    scraper = (version_scraping or {}).get(module_name, _scrape_version_attribute)
    scraped_version = scraper(module_name)

    next_module = module_name.rsplit(".", maxsplit=1)[0]
    if scraped_version is not None or next_module == module_name:
        return scraped_version
    else:
        return get_version(next_module, version_scraping=version_scraping)


def _scrape_version_attribute(module_name: str) -> str | None:
    if module_name in sys.stdlib_module_names:
        return _python_version()
    module = importlib.import_module(module_name)
    try:
        return str(module.__version__)
    except AttributeError:
        return None


def _python_version() -> str:
    vi = sys.version_info
    return f"{vi.major}.{vi.minor}.{vi.micro}"



"""
Utility functions used in pyiron.
In order to be accessible from anywhere in pyiron, they *must* remain free of any imports from pyiron!
"""

from abc import ABCMeta


class Singleton(ABCMeta):
    """
    Implemented with suggestions from

    http://stackoverflow.com/questions/6760685/creating-a-singleton-in-python

    """

    _instances: dict[type, object] = {}

    def __call__(cls, *args, **kwargs):
        if cls not in cls._instances:
            cls._instances[cls] = super().__call__(*args, **kwargs)
        return cls._instances[cls]



import os
from contextlib import contextmanager
from pathlib import Path


@contextmanager
def set_directory(path: Path | str):
    """
    A context manager that changes the current working directory to the given path
    upon entering the context and reverts to the original directory upon exiting.
    If the specified path does not exist, it is created.

    Parameters:
        path: Path | str
            The target directory path to set as the current working directory.
            If it does not exist, it will be created.

    Examples:
        Change the directory to the path "tmp" within the context:

        >>> import os
        >>> from pyiron_snippets.directory_context import set_directory
        >>> directory_before_context_is_applied = os.getcwd()
        >>> with set_directory("tmp"):
        ...     os.path.relpath(os.getcwd(), directory_before_context_is_applied)
        'tmp'

    """
    origin = Path().absolute()
    try:
        os.makedirs(path, exist_ok=True)
        os.chdir(path)
        yield
    finally:
        os.chdir(origin)



import importlib.metadata

try:
    # Installed package will find its version
    __version__ = importlib.metadata.version(__name__)
except importlib.metadata.PackageNotFoundError:
    # Repository clones will register an unknown version
    __version__ = "0.0.0+unknown"



"""
Tools for making dynamically generated classes unique, and their instances pickleable.

Provides two main user-facing tools: :func:`classfactory`, which should be used
_exclusively_ as a decorator (this restriction pertains to namespace requirements for
re-importing), and `ClassFactory`, which can be used to instantiate a new factory from
some existing factory function.

In both cases, the decorated function/input argument should be a pickleable function
taking only positional arguments, and returning a tuple suitable for use in dynamic
class creation via :func:`builtins.type` -- i.e. taking a class name, a tuple of base
classes, a dictionary of class attributes, and a dictionary of values to be expanded
into kwargs for `__subclass_init__`.

The resulting factory produces classes that are (a) pickleable, and (b) the same object
as any previously built class with the same name. (Note: avoiding class degeneracy with
respect to class name is the responsibility of the person writing the factory function.)

These classes are then themselves pickleable, and produce instances which are in turn
pickleable (so long as any data they've been fed as inputs or attributes is pickleable,
i.e. here the only pickle-barrier we resolve is that of having come from a dynamically
generated class).

Since users need to build their own class factories returning classes with sensible
names, we also provide a helper function :func:`sanitize_callable_name`, which makes
sure a string is compliant with use as a class name. This is run internally on user-
provided names, and failure for the user name and sanitized name to match will give a
clear error message.

Constructed classes can, in turn be used as bases in further class factories.
"""

from __future__ import annotations

from abc import ABCMeta
from collections.abc import Callable
from functools import wraps
from importlib import import_module
from inspect import Parameter, signature
from re import sub
from typing import ClassVar


class _SingleInstance(ABCMeta):
    """Simple singleton pattern."""

    _instance = None

    def __call__(cls, *args, **kwargs):
        if cls._instance is None:
            cls._instance = super().__call__(*args, **kwargs)
        return cls._instance


class _FactoryTown(metaclass=_SingleInstance):
    """
    Makes sure two factories created around the same factory function are the same
    factory object.
    """

    factories: dict[str, _ClassFactory] = {}

    @classmethod
    def clear(cls):
        """
        Remove factories.

        Can be useful if you're re-invoking `ClassFactory` on an updated factory
        function.
        """
        cls.factories = {}

    @staticmethod
    def _factory_address(factory_function: Callable) -> str:
        return f"{factory_function.__module__}.{factory_function.__qualname__}"

    def get_factory(
        self, factory_function: Callable[..., tuple[str, tuple[type, ...], dict, dict]]
    ) -> _ClassFactory:

        self._verify_function_only_takes_positional_args(factory_function)

        address = self._factory_address(factory_function)

        try:
            return self.factories[address]
        except KeyError:
            factory = self._build_factory(factory_function)
            self.factories[address] = factory
            return factory

    @staticmethod
    def _build_factory(factory_function):
        """
        Subclass :class:`_ClassFactory` and make an instance.
        """
        new_factory_class = type(
            sanitize_callable_name(
                f"{factory_function.__module__}{factory_function.__qualname__}"
                f"{factory_function.__name__.title()}"
                f"{_ClassFactory.__name__}"
            ).replace("_", ""),
            (_ClassFactory,),
            {},
            factory_function=factory_function,
        )
        return wraps(factory_function)(new_factory_class())

    @staticmethod
    def _verify_function_only_takes_positional_args(factory_function: Callable):
        parameters = signature(factory_function).parameters.values()
        if any(
            p.kind not in [Parameter.POSITIONAL_ONLY, Parameter.VAR_POSITIONAL]
            for p in parameters
        ):
            raise InvalidFactorySignature(
                f"{_ClassFactory.__name__} can only be subclassed using factory "
                f"functions that take exclusively positional arguments, but "
                f"{factory_function.__name__} has the parameters {parameters}"
            )


_FACTORY_TOWN = _FactoryTown()


class InvalidFactorySignature(ValueError):
    """When the factory function's arguments are not purely positional"""

    pass


class InvalidClassNameError(ValueError):
    """When a string isn't a good class name"""

    pass


class _ClassFactory(metaclass=_SingleInstance):
    """
    For making dynamically created classes the same class.
    """

    _decorated_as_classfactory: bool = False
    factory_function: ClassVar[Callable[..., tuple[str, tuple[type, ...], dict, dict]]]
    class_registry: ClassVar[dict[str, type[_FactoryMade]]] = {}

    def __init_subclass__(cls, /, factory_function, **kwargs):
        super().__init_subclass__(**kwargs)
        cls.factory_function = staticmethod(factory_function)
        cls.class_registry = {}

    def __call__(self, *args) -> type[_FactoryMade]:
        name, bases, class_dict, sc_init_kwargs = self.factory_function(*args)
        self._verify_name_is_legal(name)
        try:
            return self.class_registry[name]
        except KeyError:
            factory_made = self._build_class(
                name,
                bases,
                class_dict,
                sc_init_kwargs,
                args,
            )
            self.class_registry[name] = factory_made
            return factory_made

    @classmethod
    def clear(cls, *class_names, skip_missing=True):
        """
        Remove constructed class(es).

        Can be useful if you've updated the constructor and want to remove old
        instances.

        Args:
            *class_names (str): The names of classes to remove. Removes all of them
                when empty.
            skip_missing (bool): Whether to pass over key errors when a name is
                requested that is not currently in the class registry. (Default is
                True, let missing names pass silently.)
        """
        if len(class_names) == 0:
            cls.class_registry = {}
        else:
            for name in class_names:
                try:
                    cls.class_registry.pop(name)
                except KeyError:
                    if skip_missing:
                        continue
                    else:
                        raise KeyError(f"Could not find class {name}") from None

    def _build_class(
        self, name, bases, class_dict, sc_init_kwargs, class_factory_args
    ) -> type[_FactoryMade]:

        if "__module__" not in class_dict:
            class_dict["__module__"] = self.factory_function.__module__
        if "__qualname__" not in class_dict:
            class_dict["__qualname__"] = f"{self.factory_function.__qualname__}.{name}"
        sc_init_kwargs["class_factory"] = self
        sc_init_kwargs["class_factory_args"] = class_factory_args

        if not any(_FactoryMade in base.mro() for base in bases):
            bases = (_FactoryMade, *bases)

        return type(name, bases, class_dict, **sc_init_kwargs)

    @staticmethod
    def _verify_name_is_legal(name):
        sanitized_name = sanitize_callable_name(name)
        if name != sanitized_name:
            raise InvalidClassNameError(
                f"The class name {name} failed to match with its sanitized version"
                f"({sanitized_name}), please supply a valid class name."
            )

    def __reduce__(self):
        if (
            self._decorated_as_classfactory
            and "<locals>" not in self.factory_function.__qualname__
        ):
            return (
                _import_object,
                (self.factory_function.__module__, self.factory_function.__qualname__),
            )
        else:
            return (_FACTORY_TOWN.get_factory, (self.factory_function,))


def _import_object(module_name, qualname):
    module = import_module(module_name)
    obj = module
    for name in qualname.split("."):
        obj = getattr(obj, name)
    return obj


class _FactoryMade:
    """
    A mix-in to make class-factory-produced classes pickleable.

    If the factory is used as a decorator for another function (or class), it will
    conflict with this function (i.e. the owned function will be the true function,
    and will mismatch with imports from that location, which will return the
    post-decorator factory made class). This can be resolved by setting the
    :attr:`_reduce_imports_as` attribute to a tuple of the (module, qualname) obtained
    from the decorated definition in order to manually specify where it should be
    re-imported from. (DEPRECATED alternative: set
    :attr:`_class_returns_from_decorated_function` attribute to be the decorated
    function in the decorator definition.)
    """

    # DEPRECATED: Use _reduce_imports_as instead
    _class_returns_from_decorated_function: ClassVar[Callable | None] = None

    _reduce_imports_as: ClassVar[tuple[str, str] | None] = None  # Module and qualname

    def __init_subclass__(cls, /, class_factory, class_factory_args, **kwargs):
        super().__init_subclass__(**kwargs)
        cls._class_factory = class_factory
        cls._class_factory_args = class_factory_args
        cls._factory_town = _FACTORY_TOWN

    def __reduce__(self):
        if (
            self._class_returns_from_decorated_function is not None
            and "<locals>"
            not in self._class_returns_from_decorated_function.__qualname__
        ):
            # When we create a class by decorating some other function, this class
            # conflicts with its own factory_function attribute in the namespace, so we
            # rely on directly re-importing the factory
            return (
                _instantiate_from_decorated,
                (
                    self._class_returns_from_decorated_function.__module__,
                    self._class_returns_from_decorated_function.__qualname__,
                    self.__getnewargs_ex__(),
                ),
                self.__getstate__(),
            )
        elif (
            self._reduce_imports_as is not None
            and "<locals>" not in self._reduce_imports_as[1]
        ):
            return (
                _instantiate_from_decorated,
                (
                    self._reduce_imports_as[0],
                    self._reduce_imports_as[1],
                    self.__getnewargs_ex__(),
                ),
                self.__getstate__(),
            )
        else:
            return (
                _instantiate_from_factory,
                (
                    self._class_factory,
                    self._class_factory_args,
                    self.__getnewargs_ex__(),
                ),
                self.__getstate__(),
            )

    def __getnewargs_ex__(self):
        # Child classes can override this as needed
        return (), {}

    def __getstate__(self):
        # Python <3.11 compatibility
        try:
            return super().__getstate__()
        except AttributeError:
            return dict(self.__dict__)

    def __setstate__(self, state):
        # Python <3.11 compatibility
        try:
            super().__setstate__(state)
        except AttributeError:
            self.__dict__.update(**state)


def _instantiate_from_factory(factory, factory_args, newargs_ex):
    """
    Recover the dynamic class, then invoke its `__new__` to avoid instantiation (and
    the possibility of positional args in `__init__`).
    """
    cls = factory(*factory_args)
    return cls.__new__(cls, *newargs_ex[0], **newargs_ex[1])


def _instantiate_from_decorated(module, qualname, newargs_ex):
    """
    In case the class comes from a decorated function, we need to import it directly.
    """
    cls = _import_object(module, qualname)
    return cls.__new__(cls, *newargs_ex[0], **newargs_ex[1])


def classfactory(
    factory_function: Callable[..., tuple[str, tuple[type, ...], dict, dict]],
) -> _ClassFactory:
    """
    A decorator for building dynamic class factories whose classes are unique and whose
    terminal instances can be pickled.

    Under the hood, classes created by factories get dependence on
    :class:`_FactoryMade` mixed in. This class leverages :meth:`__reduce__` and
    :meth:`__init_subclass__` and uses up the class namespace :attr:`_class_factory`
    and :attr:`_class_factory_args` to hold data (using up corresponding public
    variable names in the :meth:`__init_subclass__` kwargs), so any interference with
    these fields may cause unexpected side effects. For un-pickling, the dynamic class
    gets recreated then its :meth:`__new__` is called using `__newargs_ex__`; a default
    implementation returning no arguments is provided on :class:`_FactoryMade` but can
    be overridden.

    Args:
        factory_function (Callable[..., tuple[str, tuple[type, ...], dict, dict]]):
            A function returning arguments that would be passed to `builtins.type` to
            dynamically generate a class. The function must accept exclusively
            positional arguments

    Returns:
        (type[_ClassFactory]): A new callable that returns unique classes whose
            instances can be pickled.

    Notes:
        If the :param:`factory_function` itself, or any data stored on instances of
        its resulting class(es) cannot be pickled, then the instances will not be able
        to be pickled. Here we only remove the trouble associated with pickling
        dynamically created classes.

        If the `__init_subclass__` kwargs are exploited, remember that these are
        subject to all the same "gotchas" as their regular non-factory use; namely, all
        child classes must specify _all_ parent class kwargs in order to avoid them
        getting overwritten by the parent class defaults!

        Dynamically generated classes can, in turn, be used as base classes for further
        `@classfactory` decorated factory functions.

    Warnings:
        Use _exclusively_ as a decorator. For an inline constructor for an existing
        callable, use :class:`ClassFactory` instead.

    Examples:
        >>> import pickle
        >>>
        >>> from pyiron_snippets.factory import classfactory
        >>> from abc import ABC
        >>>
        >>> class HasN(ABC):
        ...     '''Some class I want to make dynamically subclass.'''
        ...     def __init_subclass__(cls, /, n=0, s="foo", **kwargs):
        ...         super(HasN, cls).__init_subclass__(**kwargs)
        ...         cls.n = n
        ...         cls.s = s
        ...
        ...     def __init__(self, x, y=0):
        ...         self.x = x
        ...         self.y = y
        >>>
        >>> @classfactory
        ... def has_n_factory(n, s="wrapped_function", /):
        ...     return (
        ...         f"{HasN.__name__}{n}{s}",  # New class name
        ...         (HasN,),  # Base class(es)
        ...         {},  # Class attributes dictionary
        ...         {"n": n, "s": s}
        ...         # dict of `builtins.type` kwargs (passed to `__init_subclass__`)
        ...     )
        >>>
        >>> Has2 = has_n_factory(2, "my_dynamic_class")
        >>> HasToo = has_n_factory(2, "my_dynamic_class")
        >>> HasToo is Has2
        True

        >>> foo = Has2(42, y=-1)
        >>> print(foo.n, foo.s, foo.x, foo.y)
        2 my_dynamic_class 42 -1

        >>> reloaded = pickle.loads(pickle.dumps(foo))  # doctest: +SKIP
        >>> print(reloaded.n, reloaded.s, reloaded.x, reloaded.y)  # doctest: +SKIP
        2 my_dynamic_class 42 -1  # doctest: +SKIP

    """
    factory = _FACTORY_TOWN.get_factory(factory_function)
    factory._decorated_as_classfactory = True
    return factory


class ClassFactory:
    """
    A constructor for new class factories.

    Use on existing class factory callables, _not_ as a decorator.

    Cf. the :func:`classfactory` decorator for more info.
    """

    def __new__(cls, factory_function):
        return _FACTORY_TOWN.get_factory(factory_function)


def sanitize_callable_name(name: str):
    """
    A helper class for sanitizing a string so it's appropriate as a class/function name.
    """
    # Replace non-alphanumeric characters except underscores
    sanitized_name = sub(r"\W+", "_", name)
    # Ensure the name starts with a letter or underscore
    if (
        len(sanitized_name) > 0
        and not sanitized_name[0].isalpha()
        and sanitized_name[0] != "_"
    ):
        sanitized_name = "_" + sanitized_name
    return sanitized_name



import contextlib
from collections.abc import Callable, Collection
from typing import Any


class ExceptionExitStack(contextlib.ExitStack):
    """
    A variant of contextlib.ExitStack that only executes registered callbacks
    when an exception is raised, and only if that exception matches one of the
    specified exception types.

    Behavior:
    - If no exception types are given, callbacks run for any raised exception.
    - If one or more exception types are given, callbacks run only when the
      raised exception is an instance of at least one of those types.
    - On normal (non-exceptional) exit, callbacks are discarded and not run.
    - Exceptions are not suppressed by this context manager.

    Parameters:
        *exceptions: type[Exception]
            Zero or more exception types. If empty, callbacks run for any
            exception; otherwise, only for matching exception types.

    Examples:
        Let's take a toy callback and see how we do (or don't) trigger it.

        >>> def its_historical(history: list[str], message: str) -> None:
        ...     history.append(message)

        No types specified: callbacks run for any raised exception.

        >>> from pyiron_snippets.exception_context import ExceptionExitStack
        >>> history = []
        >>> try:
        ...     with ExceptionExitStack() as stack:
        ...         _ = stack.callback(its_historical, history, "with no types")
        ...         raise RuntimeError("Application error")
        ... except RuntimeError:
        ...     history
        ['with no types']

        Specified type(s) match(es) the raised exception: callbacks run.

        >>> history = []
        >>> try:
        ...     with ExceptionExitStack(RuntimeError) as stack:
        ...         _ = stack.callback(its_historical, history, "with matching type")
        ...         raise RuntimeError("Application error")
        ... except RuntimeError:
        ...     history
        ['with matching type']

        Specified type(s) do(es) not match the raised exception: callbacks do not run.

        >>> history = []
        >>> try:
        ...     with ExceptionExitStack(TypeError, ValueError) as stack:
        ...         _ = stack.callback(its_historical, history, "with mis-matching types")
        ...         raise RuntimeError("Application error")
        ... except RuntimeError:
        ...     history
        []

        No exception raised: callbacks do not run. But, the stack can be combined with
        other stacks.

        >>> import contextlib
        >>>
        >>> history = []
        >>> with ExceptionExitStack() as exc_stack, contextlib.ExitStack() as reg_stack:
        ...     _ = exc_stack.callback(its_historical, history, "we shouldn't see this")
        ...     _ = reg_stack.callback(its_historical, history, "but we should see this")
        >>> history
        ['but we should see this']
    """

    def __init__(self, *exceptions: type[Exception]):
        if not all(
            isinstance(e, type) and issubclass(e, Exception) for e in exceptions
        ):
            raise ValueError(
                f"Invalid exception type(s) provided. Expected only subclasses of "
                f"`Exception`, but got {exceptions}"
            )
        super().__init__()
        self._exception_types: tuple[type[Exception], ...] = (
            (Exception,) if len(exceptions) == 0 else exceptions
        )

    def __exit__(self, exc_type, exc_val, exc_tb):
        if exc_val is not None and any(
            isinstance(exc_val, e) for e in self._exception_types
        ):
            return super().__exit__(exc_type, exc_val, exc_tb)
        self.pop_all()


@contextlib.contextmanager
def on_error(
    func: Callable[..., Any],
    exceptions: type[Exception] | Collection[type[Exception]] | None,
    *args: Any,
    **kwargs: Any,
):
    """
    A context manager that invokes a callback only when an exception is raised,
    and only if that exception matches the specified type(s).

    This is analogous to ExceptionExitStack, but designed for use with an
    existing context manager stack (e.g., contextlib.ExitStack). It registers
    a single callback and defers calling it until an exception occurs and
    matches the provided exception type(s).

    Behavior:
    - If exceptions is None, the callback runs for any raised Exception.
    - If a single exception type is provided, the callback runs only when the
      raised exception is an instance of that type.
    - If a collection of exception types is provided, the callback runs when
      the raised exception matches any type in the collection.
    - On normal (non-exceptional) exit, the callback does not run.
    - Exceptions are never suppressed; they are always re-raised after the
      callback (if any) has been executed.

    Parameters:
        func: Callable[..., Any]
            The callback to execute on a matching exception.
        exceptions: type[Exception] | Collection[type[Exception]] | None
            The exception type(s) that should trigger the callback. Use None
            to match all Exceptions.
        *args: Any
            Positional arguments passed to the callback.
        **kwargs: Any
            Keyword arguments passed to the callback.

    Examples:
        A simple callback that records a message:

        >>> def its_historical(history: list[str], message: str) -> None:
        ...     history.append(message)

        Callback on all exceptions when no types are specified:

        >>> import contextlib
        >>> from pyiron_snippets.exception_context import on_error
        >>> history = []
        >>> msg = "with no types"
        >>> try:
        ...     with contextlib.ExitStack() as stack:
        ...         _ = stack.enter_context(on_error(its_historical, None, history, message=msg))
        ...         raise RuntimeError("Application error")
        ... except RuntimeError:
        ...     history
        ['with no types']

        Callback on matching exception with a specifier:

        >>> history = []
        >>> msg = "with matching type"
        >>> try:
        ...     with contextlib.ExitStack() as stack:
        ...         _ = stack.enter_context(on_error(its_historical, RuntimeError, history, message=msg))
        ...         raise RuntimeError("Application error")
        ... except RuntimeError:
        ...     history
        ['with matching type']

        No callback on mis-matching exception types:

        >>> history = []
        >>> try:
        ...     with contextlib.ExitStack() as stack:
        ...         _ = stack.enter_context(on_error(its_historical, (TypeError, ValueError), history, message="nope"))
        ...         raise RuntimeError("Application error")
        ... except RuntimeError:
        ...     history
        []

        No exception raised: callback does not run. But, we can add regular callbacks
        to the stack to combine effects.

        >>> history = []
        >>> with contextlib.ExitStack() as stack:
        ...     _ = stack.enter_context(on_error(its_historical, None, history, message="we shouldn't see this"))
        ...     _ = stack.callback(its_historical, history, message="but we should see this")
        >>> history
        ['but we should see this']
    """

    exception_types: tuple[type[Exception], ...]
    if exceptions is None:
        exception_types = (Exception,)
    elif isinstance(exceptions, type) and issubclass(exceptions, Exception):
        exception_types = (exceptions,)
    else:
        if not all(
            isinstance(e, type) and issubclass(e, Exception) for e in exceptions
        ):
            raise ValueError(
                f"Invalid exception type(s) provided. Expected only subclasses of "
                f"`Exception`, but got {exceptions}"
            )
        exception_types = tuple(exceptions)

    try:
        yield
    except Exception as e:
        if any(isinstance(e, exc_type) for exc_type in exception_types):
            func(*args, **kwargs)
        raise



"""
Classes to find data files and executables in global paths.
"""

from __future__ import annotations

import os
import os.path
import re
import warnings
from abc import ABC, abstractmethod
from collections.abc import Iterable, Iterator
from fnmatch import fnmatch
from glob import glob
from typing import Any, Self

EXE_SUFFIX = "bat" if os.name == "nt" else "sh"


class ResourceNotFound(RuntimeError):
    pass


class ResolverWarning(RuntimeWarning):
    pass


class AbstractResolver(ABC):
    """
    Interface for resolvers.

    Implementations must define :meth:`._search`, taking a tuple of names to search for and yielding instances of any
    type.  Implementations should pick a single type to yield, e.g. :class:`.ResourceResolver` always yields absolute
    paths, while :class:`.ExecutableResolver` always yields 2-tuples of a version tag and absolute paths.
    """

    @abstractmethod
    def _search(self, name: tuple[str, ...]) -> Iterator[Any]:
        pass

    def search(self, name: Iterable[str] | str = "*") -> Iterator[Any]:
        """
        Yield all matches.

        When `name` is given as an iterable, returned results match at least one of the `name` globs.

        Args:
            name (str, iterable of str): file name to search for; can be an exact file name, a glob or list of those

        Yields:
            object: resources matching `name`
        """
        if name is not None and not isinstance(name, str):
            name = tuple(name)
        else:
            name = (name,)
        yield from self._search(name)

    def list(self, name: Iterable[str] | str = "*") -> list[Any]:
        """
        Return all matches.

        Args:
            name (str, iterable of str): file name to search for; can be an exact file name, a glob or list of those

        Returns:
            list: all matches returned by :meth:`.search`.
        """
        return list(self.search(name))

    def first(self, name: Iterable[str] | str = "*") -> Any:
        """
        Return first match.

        Args:
            name (str, iterable of str): file name to search for; can be an exact file name, a glob or list of those

        Returns:
            object: the first match returned by :meth:`.search`.

        Raises:
            :class:`~.ResourceNotFound`: if no matches are found.
        """
        try:
            return next(iter(self.search(name)))
        except StopIteration:
            raise ResourceNotFound(f"Could not find {name} in {self}!") from None

    def chain(self, *resolvers: AbstractResolver) -> Self | ResolverChain:
        """
        Return a new resolver that searches this and all given resolvers sequentially.

        You will likely want to ensure that all given resolvers yield the same types and e.g. not mix ExecutableResolver
        and ResourceResolver, but this is not checked.

        The advantage of using :meth:`.chain` rather than adding more paths to one resolver is when different paths have
        different internal sub structure, such as when combining resources from pyiron resources and conda data
        packages.  When searching for lammps potential files, e.g. we have some folders that are set up as

            <resources>/lammps/potentials/...

        but iprpy conda package that ships the NIST potentials doesn't have the lammps/potentials

            <iprpy>/...

        With chaining we can do very easily

        >>> ResourceResolver([<resources>], "lammps", "potentials").chain(
        ...     ResourceResolver([<iprpy>])) # doctest: +SKIP

        without we'd need to modify the resource paths ourselves explicitly

        >>> ResourceResolver([r + '/lammps/potentials' for r in <resources>] + [<iprpy>]) # doctest: +SKIP

        which is a bit more awkward.

        Args:
            resolvers (:class:`.AbstractResolver`): any number of sub resolvers

        Returns:
            self: if `resolvers` is empty
            :class:`.ResolverChain`: otherwise
        """
        if resolvers == ():
            return self
        return ResolverChain(self, *resolvers)


class ResolverChain(AbstractResolver):
    """
    A chain of resolvers.  Matches are returned sequentially.
    """

    __slots__ = ("_resolvers",)

    def __init__(self, *resolvers):
        """
        Args:
            *resolvers (:class:`.AbstractResolver`): sub resolvers to use
        """
        self._resolvers = resolvers

    def _search(self, name):
        for resolver in self._resolvers:
            yield from resolver.search(name)

    def __repr__(self):
        inner = ", ".join(repr(r) for r in self._resolvers)
        return f"{type(self).__name__}({inner})"


class ResourceResolver(AbstractResolver):
    """
    Generic resolver for files and directories.

    Resources are expected to conform to the following format:
        <resource_path>/<module>/<subdir0>/<subdir1>/...

    *All* entries within in this final `subdir` are yielded by :meth:`.search`, whether they are files or directories.
    Search results can be restricted by passing a (list of) globs.  If a list is given, entries matching at least one of
    them are returned.

    >>> res = ResourceResolver(..., "lammps")
    >>> res.list() # doctest: +SKIP
    [
        "bin",
        "potentials",
        "potentials.csv"
    ]
    """

    __slots__ = "_resource_paths", "_module", "_subdirs"

    def __init__(self, resource_paths, module, *subdirs):
        """
        Args:
            resource_paths (list of str): base paths for resource locations
            module (str): name of the module
            *subdirs (str): additional sub directories to descend into
        """
        self._resource_paths = resource_paths
        self._module = module
        self._subdirs = subdirs

    def __repr__(self):
        inner = repr(self._resource_paths)
        inner += f", {repr(self._module)}"
        if len(self._subdirs) > 0:
            inner += ", " + ", ".join(repr(s) for s in self._subdirs)
        return f"{type(self).__name__}({inner})"

    def _search(self, name):
        for p in self._resource_paths:
            sub = os.path.join(p, self._module, *self._subdirs)
            if os.path.exists(sub):
                for n in name:
                    yield from sorted(glob(os.path.join(sub, n)))


class ExecutableResolver(AbstractResolver):
    """
    A resolver for executable scripts.

    Executables are expected to conform to the following format:
        <resource_path>/<module>/bin/run_<code>_<version_string>.<suffix>

    and have the executable bit set. :meth:`.search` yields tuples of version strings and full paths to the executable
    instead of plain strings.

    Except on windows results are filtered to make sure all returned scripts have the executable bit set.
    When the bit is not set, a warning is printed.

    >>> exe = ExecutableResolver(..., "lammps")
    >>> exe.list() # doctest: +SKIP
    [
        ('v1', '/my/resources/lammps/bin/run_lammps_v1.sh),
        ('v1_mpi', '/my/resources/lammps/bin/run_lammps_v1_mpi.sh),
        ('v2_default', '/my/resources/lammps/bin/run_lammps_v2_default.sh),
    ]
    >>> exe.default_version # doctest: +SKIP
    "v2_default"
    >>> exe.dict("v1*") # doctest: +SKIP
    {
        'v1': '/my/resources/lammps/bin/run_lammps_v1.sh),
        'v1_mpi': '/my/resources/lammps/bin/run_lammps_v1_mpi.sh)
    }
    """

    __slots__ = "_regex", "_resolver"

    def __init__(self, resource_paths, code, module=None, suffix=EXE_SUFFIX):
        """
        Args:
            resource_paths (list of str): base paths for resource locations
            code (str): name of the simulation code
            module (str): name of the module the code is part of, same as `code` by default
            suffix (str, optional): file ending; if `None`, 'bat' on Windows 'sh' elsewhere
        """
        if suffix is None:
            suffix = EXE_SUFFIX
        if module is None:
            module = code
        self._regex = re.compile(f"run_{code}_(.*)\\.{suffix}$")
        self._glob = f"run_{code}_*.{suffix}"
        self._resolver = ResourceResolver(
            resource_paths,
            module,
            "bin",
        )

    def __repr__(self):
        inner = repr(self._resolver._resource_paths)
        inner += f", {repr(self._glob)}"
        inner += f", {repr(self._resolver._module)}"
        # recover suffix
        inner += f", {repr(self._glob.split('.')[-1])}"
        return f"{type(self).__name__}({inner})"

    def _search(self, name):
        seen = set()

        def cond(path):
            isfile = os.path.isfile(path)
            # HINT: this is always True on windows
            isexec = os.access(
                path, os.X_OK, effective_ids=os.access in os.supports_effective_ids
            )
            if isfile and not isexec:
                warnings.warn(
                    f"Found file '{path}', but skipping it because it is not executable!",
                    category=ResolverWarning,
                    # TODO: maybe used from python3.12 onwards
                    # skip_file_prefixes=(os.path.dirname(__file__),),
                    stacklevel=4,
                )
            return isfile and isexec

        for path in filter(cond, self._resolver.search(self._glob)):
            # we know that the regex has to match, because we constrain the resolver with the glob
            version = self._regex.search(path).group(1)
            if version not in seen and any(fnmatch(version, n) for n in name):
                yield (version, path)
                seen.add(version)

    def dict(self, name="*") -> dict[str, str]:
        """
        Construct dict from :meth:`.search` results.

        Args:
            name (str or list of str): glob(s) to filter the version strings

        Returns:
            dict: mapping version strings to full paths
        """
        return dict(self.search(name=name))

    @property
    def available_versions(self):
        """
        list of str: all found versions
        """
        return [x[0] for x in self.search("*")]

    @property
    def default_version(self):
        """
        str: the first version found in resources

        If a version matching `*default*` exists, the first matching is returned.

        Raises:
            :class:`.ResourceNotFound`: if no executables are found at all
        """
        try:
            return self.first("*default*")[0]
        except ResourceNotFound:
            pass
        # try again outside the except clause to avoid nested error in case this fails as well
        return self.first("*")[0]



"""
Helper functions for managing the relationship between strings and imports.
"""

from __future__ import annotations

import importlib


class StringNotImportableError(ImportError): ...


def import_from_string(library_path: str) -> object:
    """
    Import an object using a string of its python library location.

    Args:
        library_path (str): The full module path to the desired object.

    Returns:
        (object): The imported object.

    Example:
        >>> from pyiron_snippets import retrieve
        >>> ThreadPoolExecutor = retrieve.import_from_string(
        ...     "concurrent.futures.ThreadPoolExecutor"
        ... )
        >>> with ThreadPoolExecutor(max_workers=2) as executor:
        ...     future = executor.submit(pow, 2, 3)
        ...     print(future.result())
        8

    """
    if (not isinstance(library_path, str)) or len(library_path) == 0:
        raise ValueError(
            f"Expected a non-empty string, got '{library_path}'  of type {type(library_path)} instead."
        )

    split_path = library_path.split(".", 1)
    if len(split_path) == 1:
        module_name, path = split_path[0], ""
    else:
        module_name, path = split_path

    try:
        obj = importlib.import_module(module_name)
    except ModuleNotFoundError as e:
        raise ModuleNotFoundError(
            f"The topmost entry of {library_path} could not be found. The most likely "
            f"causes of this problem are a typo, or that the module is not yet in your "
            f"system's PYTHONPATH. The latter can be checked from inside python with "
            f"`import sys; print(sys.path)`."
        ) from e

    for k in path.split("."):
        if k == "":
            break
        try:
            obj = getattr(obj, k)
        except AttributeError:
            # Try importing as a submodule
            # This can be necessary of an __init__.py is empty and nothing else has
            # referenced the module yet
            current_path = f"{obj.__name__}.{k}"
            obj = importlib.import_module(current_path)
    return obj


def get_importable_string_from_string_reduction(
    string_reduction: str, reduced_object: object
) -> str:
    """
    Per the pickle docs:

    > If a string is returned, the string should be interpreted as the name of a global
      variable. It should be the object’s local name relative to its module; the pickle
      module searches the module namespace to determine the object’s module. This
      behaviour is typically useful for singletons.

    To then import such an object from a non-local caller, we try scoping the string
    with the module of the object which returned it.
    """
    try:
        import_from_string(string_reduction)
        importable = string_reduction
    except ModuleNotFoundError:
        importable = reduced_object.__module__ + "." + string_reduction
        try:
            import_from_string(importable)
        except (ModuleNotFoundError, AttributeError) as e:
            raise StringNotImportableError(
                f"Couldn't import {string_reduction} after scoping it as {importable}. "
                f"Please contact the developers so we can figure out how to handle "
                f"this edge case."
            ) from e
    return importable
